#
# (C) Copyright 2014 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

from __future__ import absolute_import

from .codec import read_codec
from .constants import (
    BITFIELD,
    ENDIAN_MARKER,
    FORMAT_VERSION_NUMBER_MAJOR,
    FORMAT_VERSION_NUMBER_MINOR,
    MAGIC,
    NEW_HEADER,
    TYPE_NAMES,
)
from .stream import BigEndianStream, LittleEndianStream

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

from itertools import accumulate, chain

import pandas as pd


class MismatchedFramesError(ValueError):
    pass


class ColumnInfo:
    """
    Represent the type of a column in the encoded file

    Parameters:
        name(str): The name of the column
        idx(int): The index of the column
        dtype(DataType): The type of the column as :class:`.DataType`
        datasize(int): The size of the column
        bitfields(iter): For columns of bitfield type, define :class:`.Bitfield` specification

    Attributes:
        name(str): The name of the column
        idx(int): The index of the column
        dtype(DataType): The type of the column as :class:`.DataType`
        datasize(int): The size of the column
        bitfields(iter): For columns of bitfield type, define :class:`.Bitfield` specification
    """

    class Bitfield:
        """
        Specifies the meaning of the bits encoded in a column of type bitfield

        Parameters:
            name(str): Name of a group of bits
            size(int): Number of bits
            offset(int): Offset of bits within the decoded (integer) value

        Attributes:
            name(str): Name of a group of bits
            size(int): Number of bits
            offset(int): Offset of bits within the decoded (integer) value
        """

        def __init__(self, name, size, offset):
            self.name = name
            self.size = size
            self.offset = offset

        def __eq__(self, other):
            return self.name == other.name and self.size == other.size and self.offset == other.offset

    def __init__(self, name, idx, dtype, datasize, bitfields):
        self.name = name
        self.dtype = dtype
        self.index = idx
        self.datasize = datasize
        self.bitfields = bitfields

        assert (dtype == BITFIELD) != (len(bitfields) == 0)
        if self.bitfields:
            assert isinstance(self.bitfields, Iterable)
            assert all(isinstance(b, ColumnInfo.Bitfield) for b in self.bitfields)

    def __str__(self):
        if self.dtype == BITFIELD:
            bitfield_str = "(" + ",".join("{}:{}".format(b.name, b.size) for b in self.bitfields) + ")"
        else:
            bitfield_str = ""
        return "{}:{}{}".format(self.name, TYPE_NAMES.get(self.dtype, "<unknown>"), bitfield_str)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return (
            self.name == other.name
            and self.dtype == other.dtype
            and self.index == other.index
            and self.datasize == other.datasize  # This may be overzealous?
            and self.bitfields == other.bitfields
        )


class Frame:
    """
    Represent the decoded dataframe

    Parameters:
        source(str|file): A file-like object to decode the data from

    Attributes:
        properties(dict): Dictionary of additional properties that can contain arbitrary metadata
        columns(list): A list of :class:`.ColumnInfo` objects describing the column structure of the frame
        nrows(int): Number of rows of data within the frame
        ncolumns(int): Number of data columns within the frame
    """

    def __init__(self, source):

        # Read marker and magic

        m = source.read(2)
        if len(m) == 0:
            raise EOFError()
        assert int.from_bytes(m, byteorder="big", signed=False) == NEW_HEADER

        assert source.read(3) == MAGIC

        # Get byte ordering

        endian_marker = source.read(4)
        if int.from_bytes(endian_marker, byteorder="little") == ENDIAN_MARKER:
            stream = LittleEndianStream(source)
        else:
            stream = BigEndianStream(source)

        # TODO: Some inequalities here, rather than plain statement
        assert stream.readInt32() == FORMAT_VERSION_NUMBER_MAJOR
        assert stream.readInt32() == FORMAT_VERSION_NUMBER_MINOR

        # MD5 checksum
        assert stream.readString()

        headerLength = stream.readInt32()
        self._dataStartPosition = stream.position() + headerLength

        # 0 means we don't know offset of next header.

        self._dataSize = stream.readInt64()
        self._dataEndPosition = self._dataStartPosition + self._dataSize

        # prevFrameOffset
        assert stream.readInt64() == 0

        self._numberOfRows = stream.readInt64()

        self.flags = [stream.readReal64() for _ in range(stream.readInt32())]

        self.properties = {}

        for _ in range(stream.readInt32()):
            key = stream.readString()
            value = stream.readString()
            assert key not in self.properties
            self.properties[key] = value

        self._numberOfColumns = stream.readInt32()
        self._columnPosition = stream.position()

        self.__columnCodecs = None
        self._columns = None

        self._stream = stream

        # Support frame aggregation
        self._trailingAggregatedFrames = []

    def seekToEnd(self):
        self._stream.seek(self._dataEndPosition)

    @property
    def _column_codecs(self):
        """
        Internal method to get the codecs for the given column.
        These are read/constructed lazily from the file handle so that we can do scans through the file rapidly.

        Returns:
            list: A list of codecs
        """
        if self.__columnCodecs is None:
            self._stream.seek(self._columnPosition)
            self.__columnCodecs = [read_codec(self._stream) for _ in range(self._numberOfColumns)]
            assert self._stream.position() == self._dataStartPosition
        return self.__columnCodecs

    @property
    def columns(self):
        return [
            ColumnInfo(
                codec.column_name,
                idx,
                codec.type,
                codec.data_size,
                [
                    ColumnInfo.Bitfield(name=nm, size=sz, offset=off)
                    for nm, sz, off in zip(
                        codec.bitfield_names,
                        codec.bitfield_sizes,
                        accumulate(chain([0], codec.bitfield_sizes)),
                    )
                ],
            )
            for idx, codec in enumerate(self._column_codecs)
        ]

    @property
    def column_dict(self):
        return {c.name: c for c in self.columns}

    @property
    def simple_column_dict(self):
        return {c.name.split("@")[0]: c for c in self.columns}

    @property
    def nrows(self):
        return self._numberOfRows + sum(f.nrows for f in self._trailingAggregatedFrames)

    @property
    def ncolumns(self):
        if self.__columnCodecs is not None:
            assert self._numberOfColumns == len(self.__columnCodecs)
        return self._numberOfColumns

    def dataframe(self, columns=None):
        """
        Decodes the frame into a pandas dataframe

        Parameters:
            columns: List of columns to decode

        Returns:
            DataFrame
        """
        # TODO: Properly skip decoding columns that aren't needed
        column_codecs = self._column_codecs

        self._stream.seek(self._dataStartPosition)

        output_cols = [[] for _ in range(self._numberOfColumns)]

        # Select the correct output columns. Note we allow selection of fully-qualified
        # names, but we also allow selection of short names of the form <name>@<table> (so
        # long as these names are not ambiguous
        output = {}
        for codec, output_col in zip(column_codecs, output_cols):
            if columns is None or codec.column_name in columns:
                output[codec.column_name] = output_col
            else:
                splitname = codec.column_name.split("@")
                if len(splitname) == 2:
                    name, table = splitname
                    if name in columns:
                        if name in output:
                            raise KeyError("Ambiguous short column name '{}' requested".format(name))
                        output[name] = output_col

        lastDecoded = [0] * self._numberOfColumns

        lastStartCol = None
        for row in range(self._numberOfRows):

            startCol = self._stream.readMarker()

            if lastStartCol is None:
                if startCol > 0:
                    for col in range(startCol):
                        output_cols[col].append(column_codecs[col].typed_missing_value)

            elif lastStartCol > startCol:
                for col in range(startCol, lastStartCol):
                    last = output_cols[col][-1]
                    output_cols[col].extend(last for _ in range(row - lastDecoded[col] - 1))

            lastStartCol = startCol

            for col in range(startCol, self._numberOfColumns):
                output_cols[col].append(column_codecs[col].decode(self._stream))
                lastDecoded[col] = row

        if lastStartCol is not None:
            for col in range(lastStartCol):
                last = output_cols[col][-1]
                output_cols[col].extend(last for _ in range(self._numberOfRows - lastDecoded[col] - 1))

        df = pd.DataFrame(output)

        if len(self._trailingAggregatedFrames) > 0:
            return pd.concat(
                [df] + [f.dataframe(columns) for f in self._trailingAggregatedFrames],
                copy=False,
                axis=0,
            )
        else:
            return df

    def _append(self, frame: "Frame"):
        if self.column_dict != frame.column_dict:
            raise MismatchedFramesError
        self._trailingAggregatedFrames.append(frame)



import os
import struct

import pytest
from conftest import odc_modules

from pyodc import codec
from pyodc.constants import DataType
from pyodc.stream import LittleEndianStream


def _check_decode(cdc, encoded, check):

    st = LittleEndianStream(encoded)
    v = cdc.decode(st)
    assert v == check


def test_null_terminated_constant_string():
    """
    This tests the (somewhat dubious) 'missing' values in some (older) data
    encoded from ODB-1 using the migrator. This data uses the integer missing value,
    casted to a double, that happens to start with \x00 --> "NULL STRING"

    We need to support decoding this data...
    """
    constant_value = struct.unpack("<d", b"\x00\x00\xc0\xff\xff\xff\xdfA")[0]
    cdc = codec.ConstantString("column", constant_value, constant_value, DataType.STRING, has_missing=False)
    encoded = b""

    _check_decode(cdc, encoded, "")


def test_stripped_constant_string():
    constant_value = struct.unpack("<d", b"hello\x00\x00\x00")[0]
    cdc = codec.ConstantString("column", constant_value, constant_value, DataType.STRING, has_missing=False)
    encoded = b""

    _check_decode(cdc, encoded, "hello")


def test_normal_constant_string():
    constant_value = struct.unpack("<d", b"helloAAA")[0]
    cdc = codec.ConstantString("column", constant_value, constant_value, DataType.STRING, has_missing=False)
    encoded = b""

    _check_decode(cdc, encoded, "helloAAA")


@pytest.mark.parametrize("odyssey", odc_modules)
def test_decode_odb1_missing_strings(odyssey):
    """
    Tests that we can decode missing (NULL) strings from data encoded
    from ODB-1 using the migrator. This data uses the integer missing value,
    casted to a double, that happens to start with \x00.

    The test sample contains valid data pre-encoded (which cannot be encoded
    through the python API, as we (correctly) encode the missing value as a
    null string).
    """
    with open(os.path.join(os.path.dirname(__file__), "data/odb1_missing_string.odb"), "rb") as f:
        df = odyssey.read_odb(f, single=True)

    assert df.shape == (4, 1)
    series = df["col1"]
    assert series.dtype == "object"

    for v in series:
        assert isinstance(v, str)
        assert v == ""



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import datetime
import os
import sys

top = os.path.realpath(os.path.dirname(os.path.dirname(__file__)))
sys.path.insert(0, top)

if True:  # Keep linters happy
    import pyodc


# -- Project information -----------------------------------------------------

project = "pyodc"
author = "ECMWF"

year = datetime.datetime.now().year
if year == 2021:
    years = "2021"
else:
    years = "2021-%s" % (year,)

copyright = "%s, %s" % (years, author)

# Get the current version directly from the package.
release = pyodc.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx_rtd_theme",
    "nbsphinx",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "'**.ipynb_checkpoints'"]

source_suffix = ".rst"
master_doc = "index"
pygments_style = "sphinx"

# https://www.notion.so/Deepnote-Launch-Buttons-63c642a5e875463495ed2341e83a4b2a

nbsphinx_prolog = """
{% set docname = env.doc2path(env.docname, base=None) %}

|Binder| |Colab| |Deepnote| |Kaggle|

.. |Binder| image:: https://mybinder.org/badge.svg
   :target: https://mybinder.org/v2/gh/ecmwf/pyodc/{{ env.config.release|e }}?urlpath=lab/tree/docs/{{ docname|e }}
   :alt: Binder
   :class: badge

.. |Colab| image:: https://colab.research.google.com/assets/colab-badge.svg
   :target: https://colab.research.google.com/github/ecmwf/pyodc/blob/{{ env.config.release|e }}/docs/{{ docname|e }}
   :alt: Colab
   :class: badge

.. |Deepnote| image:: https://img.shields.io/badge/launch%20in-deepnote-blue.svg
   :target: https://beta.deepnote.org/launch?template=deepnote&url=https://github.com/ecmwf/pyodc/blob/{{ env.config.release|e }}/docs/{{ docname|e }}
   :alt: Deepnote
   :class: badge

.. |Kaggle| image:: https://kaggle.com/static/images/open-in-kaggle.svg
   :target: https://kaggle.com/kernels/welcome?src=https://github.com/ecmwf/pyodc/blob/{{ env.config.release|e }}/docs/{{ docname|e }}
   :alt: Kaggle
   :class: badge

"""  # noqa

# ipython_warning_is_error = False


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_context = {"css_files": ["_static/style.css"]}

# Remove links to the reST sources from the page headers.
html_show_sourcelink = False

# Remove "Created using Sphinx" from the HTML footer.
html_show_sphinx = False

# html_theme_options = {'logo_only': True}
# html_logo = '_static/logo.png'


# -- Options for the autodoc  extension --------------------------------------

autodoc_member_order = "bysource"


# -- Options for the todo extension ------------------------------------------

# To disable output of `todo` blocks, just set the `NO_INCLUDE_TODOS` environment variable to any value, i.e:
#
#   NO_INCLUDE_TODOS=1
#
#   Unfortunately, this cannot be done via the configuration file for the Read the Docs service, you must use their web
#   interface instead. More information here: https://github.com/readthedocs/readthedocs.org/issues/6311
todo_include_todos = False if os.environ.get("NO_INCLUDE_TODOS") else True


# -- Options for the intersphinx extension ------------------------------------------

# Target the latest version when interlinking the projects
intersphinx_mapping = {
    "odc": ("https://odc.readthedocs.io/en/latest/", None),
}



import os

import pytest

import pyodc

odc_modules = [pyodc]

codc = None
ODC_VERSION = None

if "PYODC_SKIP_CODC" not in os.environ:
    try:
        import codc

        odc_modules.append(codc)

        tmp_str = codc.lib.ffi.new("char**")
        codc.lib.lib.odc_version(tmp_str)
        ODC_VERSION = codc.lib.ffi.string(tmp_str[0]).decode("utf-8")

    except ImportError:
        msg = "Could not import codc, do you have odc installed?\nTo skip codc tests, set PYODC_SKIP_CODC flag."
        pytest.exit(msg)


def pytest_terminal_summary(terminalreporter):
    """Print out a warning if the codc tests were skipped"""

    if not codc:
        terminalreporter.ensure_newline()
        terminalreporter.section("WARNING", sep="-", yellow=True, bold=True)
        terminalreporter.line("PYODC_SKIP_CODC flag set, codc tests were skipped", yellow=True)



#!/usr/bin/env python3

import sys

from pycparser import c_generator, parse_file


def usage():
    sys.stderr.write("Usage:\n")
    sys.stderr.write("    build_header.py <odc.h> <odc_cffi.h>\n")


if len(sys.argv) != 3:
    usage()
    sys.exit(-1)

input_filename = sys.argv[1]
output_filename = sys.argv[2]

ast = parse_file(input_filename, use_cpp=True)
with open(output_filename, "w") as f:
    f.write(c_generator.CGenerator().visit(ast))



from tempfile import NamedTemporaryFile

import numpy.testing
import pandas
import pytest
from conftest import ODC_VERSION, codc, odc_modules

SAMPLE_DATA = {
    "col1": [1, 2, 3, 4, 5, 6, 7],
    "col2": [0, 0, 0, 0, 0, 0, 0],
    "col3": [73] * 7,
    "col4": [1.432] * 7,
    "col5": [-17, -7, -7, None, 1, 4, 4],
    "col6": ["aoeu", "aoeu", "aaaaaaaooooooo", "None", "boo", "squiggle", "a"],
    "col7": ["abcd"] * 7,
    "col8": [2.345] * 7,
    "col9": [999.99, 888.88, 777.77, 666.66, 555.55, 444.44, 333.33],
    "col10": [999.99, 888.88, 777.77, 666.66, 555.55, 444.44, 333.33],
    "col11": [1, None, 3, 4, 5, None, 7],
    "col12": [-512, None, 3, 7623, -22000, None, 7],
    "col13": [-1234567, 8765432, None, 22, 22222222, -81222323, None],
    # 'col21': [None] * 7
}

SAMPLE_PROPERTIES = {
    "property1": "this is a string ....",
    "property2": ".......and another .......",
}


def encode_sample(odyssey, f):

    df = pandas.DataFrame(SAMPLE_DATA)

    types = {
        "col8": odyssey.REAL,
        "col10": odyssey.REAL,
        # 'col21': odyssey.REAL
    }

    properties = SAMPLE_PROPERTIES

    odyssey.encode_odb(df, f, types=types, rows_per_frame=4, properties=properties)

    if not isinstance(f, str):
        f.flush()

    return df


@pytest.mark.parametrize("odyssey", odc_modules)
def test_encode_decode_filename(odyssey):

    with NamedTemporaryFile() as fencode:
        df = encode_sample(odyssey, fencode)

        df2 = odyssey.read_odb(fencode.name, single=True)
        assert isinstance(df2, pandas.DataFrame)

        for col in df.keys():
            s1 = df[col]
            s2 = df2[col]
            if col in ("col8", "col10"):
                numpy.testing.assert_array_almost_equal(s1, s2, decimal=2)
            else:
                numpy.testing.assert_array_equal(s1, s2)


@pytest.mark.parametrize("odyssey", odc_modules)
def test_encode_decode_file_object(odyssey):

    with NamedTemporaryFile() as fencode:
        df = encode_sample(odyssey, fencode)

        with open(fencode.name, "rb") as fread:
            df2 = odyssey.read_odb(fread, single=True)
            assert isinstance(df2, pandas.DataFrame)

        for col in df.keys():
            s1 = df[col]
            s2 = df2[col]
            if col in ("col8", "col10"):
                numpy.testing.assert_array_almost_equal(s1, s2, decimal=2)
            else:
                numpy.testing.assert_array_equal(s1, s2)


@pytest.mark.parametrize("odyssey", odc_modules)
def test_encode_decode_simple_columns(odyssey):

    with NamedTemporaryFile() as fencode:
        df = encode_sample(odyssey, fencode)

        cols = ("col6", "col7")
        df2 = odyssey.read_odb(fencode.name, columns=cols, single=True)
        assert isinstance(df2, pandas.DataFrame)
        assert df2.shape[1] == len(cols)

        for col in cols:
            numpy.testing.assert_array_equal(df[col], df2[col])


@pytest.mark.parametrize("odyssey", odc_modules)
def test_aggregate_non_matching(odyssey):
    """
    Where we aggregate tables with non-matching columns, ensure that the infilled
    missing values are type appropriate
    """
    sample1 = {"col1": [111, 222, 333]}

    sample2 = {"col2": ["aaa", "bbb", "ccc"]}

    with NamedTemporaryFile() as fencode:

        odyssey.encode_odb(pandas.DataFrame(sample1), fencode)
        odyssey.encode_odb(pandas.DataFrame(sample2), fencode)
        fencode.flush()

        df = odyssey.read_odb(fencode.name, single=True)
        assert isinstance(df, pandas.DataFrame)

        assert df["col1"].dtype == "float64"
        assert df["col2"].dtype == "object"
        assert df["col2"][0] is None
        numpy.testing.assert_array_equal(df["col1"], [111.0, 222.0, 333.0, numpy.nan, numpy.nan, numpy.nan])
        numpy.testing.assert_array_equal(df["col2"], [None, None, None, "aaa", "bbb", "ccc"])


@pytest.mark.parametrize("odyssey", odc_modules)
def test_unqualified_names(odyssey):
    """
    Check that we can extract columns by unqualified name, and by fully qualified name
    (that is, where the name contains an '@', we can extract by short name
    """
    sample = {
        "col1@tbl1": [11, 12, 13, 14, 15, 16],
        "col2@tbl1": [21, 22, 23, 24, 25, 26],
        "col1@tbl2": [31, 32, 33, 34, 35, 36],
        "col3@tbl2": [41, 42, 43, 44, 45, 46],
        "col4": [51, 52, 53, 54, 55, 56],
    }

    input_df = pandas.DataFrame(sample)

    with NamedTemporaryFile() as fencode:
        odyssey.encode_odb(input_df, fencode)
        fencode.flush()

        # Check fully-qualified naming

        cols = ["col1@tbl1", "col3@tbl2", "col4"]
        df = odyssey.read_odb(fencode.name, single=True, columns=cols)
        expected_df = input_df[cols]
        assert expected_df.equals(df)

        # Check quick-access naming

        cols = ["col2", "col3", "col4"]
        df = odyssey.read_odb(fencode.name, single=True, columns=cols)
        assert len(df.keys()) == 3
        assert set(df.keys()) == set(cols)
        expected_df = pandas.DataFrame(
            {
                "col2": sample["col2@tbl1"],
                "col3": sample["col3@tbl2"],
                "col4": sample["col4"],
            }
        )
        assert expected_df.equals(df)

        # What happens if we try and access an ambiguous columns?

        with pytest.raises(KeyError):
            odyssey.read_odb(fencode.name, single=True, columns=["col1"])


@pytest.mark.parametrize("odyssey", odc_modules)
def test_encode_decode_properties(odyssey):
    """Check that additional properties are encoded and decoded properly."""

    with NamedTemporaryFile() as fencode:

        # Test both by passing file handle and name to the encoding function.
        #   Please see ODB-523 for more information.
        for f in [fencode, fencode.name]:
            encode_sample(odyssey, f)

            reader = odyssey.Reader(fencode.name)

            for frame in reader.frames:
                expected_properties = SAMPLE_PROPERTIES

                # C library adds an implicit `encoder` property to each frame
                if odyssey == codc:
                    expected_properties["encoder"] = "odc version " + ODC_VERSION

                assert frame.properties == expected_properties



import struct
from enum import IntEnum, unique

# Current version information

FORMAT_VERSION_NUMBER_MAJOR = 0
FORMAT_VERSION_NUMBER_MINOR = 5

# Default missing values

MISSING_INTEGER = 2147483647
MISSING_REAL = -2147483647
MISSING_STRING = struct.unpack("<d", b"\x00\x00\x00\x00\x00\x00\x00\x00")[0]
DECODED_MISSING_STRING = ""

INTERNAL_REAL_MISSING = (
    struct.unpack("<f", b"\x00\x00\x80\x00")[0],
    struct.unpack("<f", b"\xff\xff\x7f\x7f")[0],
)

# Constants for header encoding

NEW_HEADER = 65535
MAGIC = b"ODA"
ENDIAN_MARKER = 1

# Types


@unique
class DataType(IntEnum):
    """
    Defines the encoded data type for a specified column
    """

    IGNORE = 0  #: Ignore value
    INTEGER = 1  #: Integer value
    REAL = 2  #: Real value
    STRING = 3  #: String value
    BITFIELD = 4  #: Bitfield value
    DOUBLE = 5  #: Double value


IGNORE = DataType.IGNORE
INTEGER = DataType.INTEGER
REAL = DataType.REAL
STRING = DataType.STRING
BITFIELD = DataType.BITFIELD
DOUBLE = DataType.DOUBLE

TYPE_NAMES = {
    IGNORE: "ignore",
    INTEGER: "integer",
    REAL: "real",
    STRING: "string",
    BITFIELD: "bitfield",
    DOUBLE: "double",
}



# This tests a specific case of optimisation from the old encoder, WriteBufferingIterator.
#
# This pre-initialised the rows buffer with missing values. As such, for the first row of
# data, if these contained missing values, it did not differ from the previous row.
#
# If the first values in the row to be encoded (i.e. the slowest varying column) contained
# a missing value, then according to this metric, it did not need to be encoded, and the
# starting row marker would be non-zero.
#
# We need to ensure that these get correctly decoded!

# See ODB-533

import os

import numpy
import pandas
import pytest
from conftest import odc_modules


@pytest.mark.parametrize("odyssey", odc_modules)
def test_initial_missing1(odyssey):

    check_data = {
        "stringval": ["", "testing"],
        "intval": [None, 12345678],
        "realval": [None, 1234.56],
        "doubleval": [None, 9876.54],
    }

    check_df = pandas.DataFrame(check_data)

    data_file = os.path.join(os.path.dirname(__file__), "data/odb_533_1.odb")
    df = odyssey.read_odb(data_file, single=True)

    print(check_df)
    print(df)

    assert set(check_df.columns) == set(df.columns)
    for col in df.columns:
        numpy.testing.assert_array_equal(df[col], check_df[col])


@pytest.mark.parametrize("odyssey", odc_modules)
def test_initial_missing2(odyssey):

    check_data = {
        "stringval": ["", "testing"],
        "intval": [None, 12345678],
        "realval": [None, 1234.56],
        "doubleval": [None, 9876.54],
        "changing": [1234, 5678],
    }

    check_df = pandas.DataFrame(check_data)

    data_file = os.path.join(os.path.dirname(__file__), "data/odb_533_2.odb")
    df = odyssey.read_odb(data_file, single=True)

    print(check_df)
    print(df)

    assert set(check_df.columns) == set(df.columns)
    for col in df.columns:
        numpy.testing.assert_array_equal(df[col], check_df[col])



import io

import pandas as pd
import pytest

from pyodc import codec
from pyodc.codec import select_codec
from pyodc.stream import LittleEndianStream


def _check_encode(codec, series, encode_compare):

    f = io.BytesIO()
    st = LittleEndianStream(f)

    for v in series:
        codec.encode(st, v)

    f.seek(0)
    assert f.read() == encode_compare


def test_int8_range_encoding():

    # Also test with negative numbers!

    for offset in (0, -100):

        s = pd.Series((1 + offset, 2 ** 8 + offset))
        c = select_codec("column", s, None)

        assert type(c) == codec.Int8
        assert c.min == 1 + offset

    _check_encode(c, s, b"\x00\xff")


def test_int16_range_encoding_minimal():
    """
    A span of integers that _just_ requires int16
    """

    # Also test with negative numbers!

    for offset in (0, -10000):

        s = pd.Series((1 + offset, 2 ** 8 + offset + 1))
        c = select_codec("column", s, None)

        assert type(c) == codec.Int16
        assert c.min == 1 + offset

        _check_encode(c, s, b"\x00\x00\x00\x01")


def test_int16_range_encoding_maximal():

    # Also test with negative numbers!

    for offset in (0, -10000):

        s = pd.Series((1 + offset, 2 ** 8 + offset, 2 ** 16 + offset))
        c = select_codec("column", s, None)

        assert type(c) == codec.Int16
        assert c.min == 1 + offset

        _check_encode(c, s, b"\x00\x00\xff\x00\xff\xff")


def test_int32_range_encoding():
    """
    n.b. the Int32 codec is a bit crappy. It does _not_ include an offset value
    --> It only encodes the legit values of a SIGNED 32bit integer
    --> 64bit integers are todo (but break some fortran compatibility, as not all
        64bit integers can be represented as doubles).
    --> Can include missing values
    """
    s = pd.Series((-(2 ** 31), None, 2 ** 31 - 2))
    c = select_codec("column", s, None)

    assert isinstance(c, codec.Int32)
    assert c.min == -(2 ** 31)

    _check_encode(c, s, b"\x00\x00\x00\x80\xff\xff\xff\x7f\xfe\xff\xff\x7f")


def test_wider_range_unsupported():
    s = pd.Series((-(2 ** 31), 2 ** 31 - 1))
    with pytest.raises(NotImplementedError):
        select_codec("column", s, None)


def test_int8_missing_range_encoding():

    # Also test with negative numbers!

    for offset in (0, -100):

        s = pd.Series((1 + offset, None, 2 ** 8 + offset - 1))
        c = select_codec("column", s, None)

        assert type(c) == codec.Int8Missing
        assert c.min == 1 + offset

    _check_encode(c, s, b"\x00\xff\xfe")


def test_int16_missing_range_encoding_minimal():

    # Also test with negative numbers!

    for offset in (0, -100):

        s = pd.Series((1 + offset, None, 2 ** 8 + offset))
        c = select_codec("column", s, None)

        assert type(c) == codec.Int16Missing
        assert c.min == 1 + offset

    _check_encode(c, s, b"\x00\x00\xff\xff\xff\x00")


def test_int16_missing_range_encoding_maximal():

    # Also test with negative numbers!

    for offset in (0, -100):

        s = pd.Series((1 + offset, None, 2 ** 16 + offset - 1))
        c = select_codec("column", s, None)

        assert type(c) == codec.Int16Missing
        assert c.min == 1 + offset

    _check_encode(c, s, b"\x00\x00\xff\xff\xfe\xff")



import os

import pytest
from conftest import codc, odc_modules


def exception_map(module, exception):
    if not codc:
        return exception
    return codc.ODCException if module == codc else exception


data_file1 = os.path.join(os.path.dirname(__file__), "data/data1.odb")

data_file1_bitfields = {
    "report_status@hdr": [
        {"name": "active", "offset": 0, "size": 1},
        {"name": "passive", "offset": 1, "size": 1},
        {"name": "rejected", "offset": 2, "size": 1},
        {"name": "blacklisted", "offset": 3, "size": 1},
        {"name": "use_emiskf_only", "offset": 4, "size": 1},
    ],
    "report_event1@hdr": [
        {"name": "no_data", "offset": 0, "size": 1},
        {"name": "all_rejected", "offset": 1, "size": 1},
        {"name": "bad_practice", "offset": 2, "size": 1},
        {"name": "rdb_rejected", "offset": 3, "size": 1},
        {"name": "redundant", "offset": 4, "size": 1},
        {"name": "stalt_missing", "offset": 5, "size": 1},
        {"name": "qc_failed", "offset": 6, "size": 1},
        {"name": "overcast_ir", "offset": 7, "size": 1},
        {"name": "thinned", "offset": 8, "size": 1},
        {"name": "latlon_corrected", "offset": 9, "size": 1},
        {"name": "stalt_corrected", "offset": 10, "size": 1},
    ],
    "report_rdbflag@hdr": [
        {"name": "lat_humon", "offset": 0, "size": 1},
        {"name": "lat_qcsub", "offset": 1, "size": 1},
        {"name": "lat_override", "offset": 2, "size": 1},
        {"name": "lat_flag", "offset": 3, "size": 2},
        {"name": "lat_hqc_flag", "offset": 5, "size": 1},
        {"name": "lon_humon", "offset": 6, "size": 1},
        {"name": "lon_qcsub", "offset": 7, "size": 1},
        {"name": "lon_override", "offset": 8, "size": 1},
        {"name": "lon_flag", "offset": 9, "size": 2},
        {"name": "lon_hqc_flag", "offset": 11, "size": 1},
        {"name": "date_humon", "offset": 12, "size": 1},
        {"name": "date_qcsub", "offset": 13, "size": 1},
        {"name": "date_override", "offset": 14, "size": 1},
        {"name": "date_flag", "offset": 15, "size": 2},
        {"name": "date_hqc_flag", "offset": 17, "size": 1},
        {"name": "time_humon", "offset": 18, "size": 1},
        {"name": "time_qcsub", "offset": 19, "size": 1},
        {"name": "time_override", "offset": 20, "size": 1},
        {"name": "time_flag", "offset": 21, "size": 2},
        {"name": "time_hqc_flag", "offset": 23, "size": 1},
        {"name": "stalt_humon", "offset": 24, "size": 1},
        {"name": "stalt_qcsub", "offset": 25, "size": 1},
        {"name": "stalt_override", "offset": 26, "size": 1},
        {"name": "stalt_flag", "offset": 27, "size": 2},
        {"name": "stalt_hqc_flag", "offset": 29, "size": 1},
    ],
    "level@conv_body": [
        {"name": "maxwind", "offset": 0, "size": 1},
        {"name": "tropopause", "offset": 1, "size": 1},
        {"name": "d_part", "offset": 2, "size": 1},
        {"name": "c_part", "offset": 3, "size": 1},
        {"name": "b_part", "offset": 4, "size": 1},
        {"name": "a_part", "offset": 5, "size": 1},
        {"name": "surface", "offset": 6, "size": 1},
        {"name": "signwind", "offset": 7, "size": 1},
        {"name": "signtemp", "offset": 8, "size": 1},
    ],
    "datum_anflag@body": [
        {"name": "final", "offset": 0, "size": 4},
        {"name": "fg", "offset": 4, "size": 4},
        {"name": "depar", "offset": 8, "size": 4},
        {"name": "varqc", "offset": 12, "size": 4},
        {"name": "blacklist", "offset": 16, "size": 4},
        {"name": "ups", "offset": 20, "size": 1},
        {"name": "uvt", "offset": 21, "size": 1},
        {"name": "uhu", "offset": 22, "size": 1},
        {"name": "ut2", "offset": 23, "size": 1},
        {"name": "uh2", "offset": 24, "size": 1},
        {"name": "uv1", "offset": 25, "size": 1},
        {"name": "urr", "offset": 26, "size": 1},
        {"name": "usn", "offset": 27, "size": 1},
        {"name": "usst", "offset": 28, "size": 1},
    ],
    "datum_status@body": [
        {"name": "active", "offset": 0, "size": 1},
        {"name": "passive", "offset": 1, "size": 1},
        {"name": "rejected", "offset": 2, "size": 1},
        {"name": "blacklisted", "offset": 3, "size": 1},
        {"name": "use_emiskf_only", "offset": 4, "size": 1},
    ],
    "datum_event1@body": [
        {"name": "vertco_missing", "offset": 0, "size": 1},
        {"name": "obsvalue_missing", "offset": 1, "size": 1},
        {"name": "fg_missing", "offset": 2, "size": 1},
        {"name": "rdb_rejected", "offset": 3, "size": 1},
        {"name": "assim_cld_flag", "offset": 4, "size": 1},
        {"name": "bad_practice", "offset": 5, "size": 1},
        {"name": "vertpos_outrange", "offset": 6, "size": 1},
        {"name": "fg2big", "offset": 7, "size": 1},
        {"name": "depar2big", "offset": 8, "size": 1},
        {"name": "obs_error2big", "offset": 9, "size": 1},
        {"name": "datum_redundant", "offset": 10, "size": 1},
        {"name": "level_redundant", "offset": 11, "size": 1},
        {"name": "not_analysis_varno", "offset": 12, "size": 1},
        {"name": "duplicate", "offset": 13, "size": 1},
        {"name": "levels2many", "offset": 14, "size": 1},
        {"name": "level_selection", "offset": 15, "size": 1},
        {"name": "vertco_consistency", "offset": 16, "size": 1},
        {"name": "vertco_type_changed", "offset": 17, "size": 1},
        {"name": "combined_flagging", "offset": 18, "size": 1},
        {"name": "report_rejected", "offset": 19, "size": 1},
        {"name": "varqc_performed", "offset": 20, "size": 1},
        {"name": "obserror_increased", "offset": 21, "size": 1},
        {"name": "contam_cld_flag", "offset": 22, "size": 1},
        {"name": "contam_rain_flag", "offset": 23, "size": 1},
        {"name": "contam_aerosol_flag", "offset": 24, "size": 1},
        {"name": "bad_emissivity", "offset": 25, "size": 1},
        {"name": "model_cld_flag", "offset": 26, "size": 1},
    ],
    "datum_rdbflag@body": [
        {"name": "press_humon", "offset": 0, "size": 1},
        {"name": "press_qcsub", "offset": 1, "size": 1},
        {"name": "press_override", "offset": 2, "size": 1},
        {"name": "press_flag", "offset": 3, "size": 2},
        {"name": "press_hqc_flag", "offset": 5, "size": 1},
        {"name": "press_judged_prev_an", "offset": 6, "size": 2},
        {"name": "press_used_prev_an", "offset": 8, "size": 1},
        {"name": "_press_unused_6", "offset": 9, "size": 6},
        {"name": "varno_humon", "offset": 15, "size": 1},
        {"name": "varno_qcsub", "offset": 16, "size": 1},
        {"name": "varno_override", "offset": 17, "size": 1},
        {"name": "varno_flag", "offset": 18, "size": 2},
        {"name": "varno_hqc_flag", "offset": 20, "size": 1},
        {"name": "varno_judged_prev_an", "offset": 21, "size": 2},
        {"name": "varno_used_prev_an", "offset": 23, "size": 1},
    ],
}


@pytest.mark.parametrize("odyssey", odc_modules)
def test_error_handling(odyssey):
    """Ensure that exceptions in the C++ layer are caught and forwarded as python exceptions"""
    with pytest.raises(exception_map(odyssey, FileNotFoundError)):
        odyssey.Reader("No such ODB file")


@pytest.mark.parametrize("odyssey", odc_modules)
def test_frame_column_bitfields(odyssey):
    """
    Check if bitfield columns are properly described

    Please see ODB-523 for more information.
    """

    r = odyssey.Reader(data_file1)

    for frame in r.frames:
        for column_name in data_file1_bitfields:
            column = frame.column_dict[column_name]

            assert column.dtype == odyssey.DataType.BITFIELD
            assert len(column.bitfields) == len(data_file1_bitfields[column.name])

            for bf, bf_expected in zip(column.bitfields, data_file1_bitfields[column.name]):
                assert bf.name == bf_expected["name"]
                assert bf.size == bf_expected["size"]
                assert bf.offset == bf_expected["offset"]



from .constants import BITFIELD, DOUBLE, IGNORE, INTEGER, REAL, STRING, DataType
from .encoder import encode_odb
from .frame import ColumnInfo, Frame
from .lib import ODCException
from .reader import Reader, read_odb

__version__ = "1.1.2"



import os
import subprocess
import sys


def test_setup():
    """Performs some tests on the meta-data of a package"""
    repodir = os.path.dirname(os.path.dirname(__file__))
    setup_py = os.path.join(repodir, "setup.py")
    subprocess.check_call([sys.executable, setup_py, "check"])



import os

import pytest
from conftest import odc_modules

data_file1 = os.path.join(os.path.dirname(__file__), "data/data1.odb")


@pytest.mark.parametrize("odyssey", odc_modules)
def test_count(odyssey):

    # print("We are here")
    # print(odyssey.__version__)

    r = odyssey.Reader(data_file1, aggregated=False)

    frames = r.frames

    assert len(frames) == 11
    assert all(t.ncolumns == 55 for t in frames)
    assert sum(t.nrows for t in frames) == 13


@pytest.mark.parametrize("odyssey", odc_modules)
def test_count_file(odyssey):
    """Check that we can open a file like object"""
    with open(data_file1, "rb") as f:
        r = odyssey.Reader(f, aggregated=False)

    frames = r.frames
    assert len(frames) == 11
    assert all(t.ncolumns == 55 for t in frames)
    assert sum(t.nrows for t in frames) == 13


@pytest.mark.parametrize("odyssey", odc_modules)
def test_count_aggregated_file(odyssey):
    """Check that we can open a file like object"""
    with open(data_file1, "rb") as f:
        r = odyssey.Reader(f, aggregated=True)

    frames = r.frames
    assert len(frames) == 1
    assert all(t.ncolumns == 55 for t in frames)
    assert sum(t.nrows for t in frames) == 13


if __name__ == "__main__":
    pytest.main()



import io

import pandas

from .frame import Frame, MismatchedFramesError


class Reader:
    """
    An object that owns the input data stream, and splits it into a sequence of frames that can be interrogated

    Parameters:
        source(str|file): A file-like object to decode the data from
        aggregated(bool): Group result into logical dataframes if ``True``

    Attributes:
        frames(DataFrame): Decoded dataframes
    """

    def __init__(self, source, aggregated=True):
        self.__aggregated = aggregated
        self._frames = []

        if isinstance(source, io.IOBase):
            self._f = source
        else:
            self._f = open(source, "rb")

        while True:
            try:
                self._frames.append(Frame(self._f))
                self._frames[-1].seekToEnd()
            except EOFError:
                # n.b. f.read() does not throw EOFError, so this will only catch the exception
                # thrown internally when the table marker is not found
                break

        if self.__aggregated:
            if len(self._frames) > 1:
                aggregated_frames = [self._frames[0]]
                for frame in self._frames[1:]:
                    try:
                        aggregated_frames[-1]._append(frame)
                    except MismatchedFramesError:
                        aggregated_frames.append(frame)
                self._frames = aggregated_frames

    @property
    def frames(self):
        return self._frames


def _read_odb_generator(source, columns=None, aggregated=True):
    r = Reader(source, aggregated=aggregated)
    for f in r.frames:
        yield f.dataframe(columns)


def _read_odb_oneshot(source, columns=None):
    reduced = pandas.concat(_read_odb_generator(source, columns), sort=False, ignore_index=True)
    for name, data in reduced.iteritems():
        if data.dtype == "object":
            data.where(pandas.notnull(data), None, inplace=True)
    return reduced


def read_odb(source, columns=None, aggregated=True, single=False):
    """
    Decode an ODB-2 stream into a pandas dataframe

    Parameters:
        source(str|file): A file-like object to decode the data from
        columns(list|tuple): A list or a tuple of columns to decode
        aggregated(bool): Group result into logical dataframes if ``True``
        single(bool): Group result into a single dataframe if ``True`` and possible

    Returns:
        DataFrame
    """
    if single:
        assert aggregated
        return _read_odb_oneshot(source, columns)
    else:
        return _read_odb_generator(source, columns, aggregated)



import io
import os
import re

import setuptools


def read(path):
    file_path = os.path.join(os.path.dirname(__file__), *path.split("/"))
    return io.open(file_path, encoding="utf-8").read()


# single-sourcing the package version using method 1 of:
#   https://packaging.python.org/guides/single-sourcing-package-version/
def parse_version_from(path):
    version_file = read(path)
    version_match = re.search(r'^__version__ = "(.*)"', version_file, re.M)
    if version_match is None or len(version_match.groups()) > 1:
        raise ValueError("couldn't parse version")
    return version_match.group(1)


setuptools.setup(
    name="pyodc",
    version=parse_version_from("codc/__init__.py"),
    description="A Python interface to odc for encoding/decoding ODB-2 files.",
    long_description=read("README.md") + read("CHANGELOG.md"),
    long_description_content_type="text/markdown",
    author="European Centre for Medium-Range Weather Forecasts (ECMWF)",
    author_email="software.support@ecmwf.int",
    license="Apache License Version 2.0",
    url="https://github.com/ecmwf/pyodc",
    packages=["pyodc", "codc"],
    package_data={"": ["*.h"]},
    include_package_data=True,
    install_requires=[
        "cffi",
        "pandas",
    ],
    extras_require={},
    tests_require=[
        "pytest",
        "pytest-cov",
        "pytest-flakes",
    ],
    test_suite="tests",
    zip_safe=True,
    keywords="odc odb",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Operating System :: OS Independent",
    ],
)
